# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import logging

import boto.s3
log = logging.getLogger( __name__ )

old_match_hostname = None


def work_around_dots_in_bucket_names( ):
    """
    https://github.com/boto/boto/issues/2836
    """
    global old_match_hostname
    if old_match_hostname is None:
        import re
        import ssl
        try:
            old_match_hostname = ssl.match_hostname
        except AttributeError:
            log.warn( 'Failed to install workaround for dots in bucket names.' )
        else:
            hostname_re = re.compile( r'^(.*?)(\.s3(?:-[^.]+)?\.amazonaws\.com)$' )

            def new_match_hostname( cert, hostname ):
                match = hostname_re.match( hostname )
                if match:
                    hostname = match.group( 1 ).replace( '.', '' ) + match.group( 2 )
                return old_match_hostname( cert, hostname )

            ssl.match_hostname = new_match_hostname


def region_to_bucket_location( region ):
    if region == 'us-east-1':
        return ''
    else:
        return region


def bucket_location_to_region( location ):
    if location == '':
        return 'us-east-1'
    else:
        return location


def s3_connect_to_region( region ):
    s3 = boto.s3.connect_to_region( region )
    if s3 is None:
        raise RuntimeError( "The region name '%s' appears to be invalid.", region )
    else:
        return s3


def modify_metadata_retry( ):
    """
    https://github.com/BD2KGenomics/s3am/issues/16
    """
    from boto import config

    def inject_default( name, default ):
        section = 'Boto'
        value = config.get( section, name )

        if value != default:
            if not config.has_section( section ):
                config.add_section( section )
            config.set( section, name, default )

    inject_default( 'metadata_service_timeout', '5.0' )
    inject_default( 'metadata_service_num_attempts', '3' )



# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import base64
import functools
import random
from contextlib import closing, contextmanager
import hashlib
from operator import itemgetter
from bd2k.util.ec2.credentials import enable_metadata_credential_caching

import time

import abc
import os
import pycurl
import multiprocessing
import itertools
from io import BytesIO
import signal
import traceback
from urlparse import urlparse
from StringIO import StringIO

from boto import handler
from boto.resultset import ResultSet
from boto.s3.connection import S3Connection, xml

from boto.s3.multipart import MultiPartUpload, Part

from s3am import me, log, UserError, WorkerException
from s3am.boto_utils import (work_around_dots_in_bucket_names,
                             s3_connect_to_region,
                             bucket_location_to_region)
from s3am.humanize import bytes2human, human2bytes

max_part_per_page = 1000  # http://docs.aws.amazon.com/AmazonS3/latest/dev/qfacts.html
max_uploads_per_page = 1000
min_part_size = human2bytes( "5M" )
max_part_size = human2bytes( "5G" )
max_parts_per_upload = 10000

verify_buffer_size = 1024 * 1024

# The multiprocessing module needs semaphores to be declared at the top level. I'm assuming this
# applies to events, too. The semaphore's initial value depends on a command line option so we
# need to instantiate it later, in stream(). The second reason for initializing them later is
# that we want stream() to be called multiple times per program invocation, e.g. for unit tests.
#
download_slots_semaphore = None
done_event = None
error_event = None

num_cores = multiprocessing.cpu_count( )


class Operation( object ):
    """
    A S3AM operation
    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def run( self ):
        raise NotImplementedError( )

    def __init__( self, requester_pays=False ):
        super( Operation, self ).__init__( )
        self.requester_pays = requester_pays
        work_around_dots_in_bucket_names( )
        enable_metadata_credential_caching( )

    @staticmethod
    def _add_encryption_headers( sse_key, headers, for_copy=False ):
        assert len( sse_key ) == 32
        encoded_sse_key = base64.b64encode( sse_key )
        encoded_sse_key_md5 = base64.b64encode( hashlib.md5( sse_key ).digest( ) )
        prefix = 'x-amz%s-server-side-encryption-customer-' % ('-copy-source' if for_copy else '')
        headers[ prefix + 'algorithm' ] = 'AES256'
        headers[ prefix + 'key' ] = encoded_sse_key
        headers[ prefix + 'key-md5' ] = encoded_sse_key_md5

    def _get_default_headers( self ):
        headers = { }
        if self.requester_pays:
            headers[ 'x-amz-request-payer' ] = 'requester'
        return headers

    # Work around bucket.get_location() not supporting the headers keyword argument

    def get_bucket_location( self, bucket, headers=None ):
        response = bucket.connection.make_request( 'GET', bucket.name,
                                                   query_args='location', headers=headers )
        body = response.read( )
        if response.status == 200:
            rs = ResultSet( bucket )
            h = handler.XmlHandler( rs, bucket )
            if not isinstance( body, bytes ):
                body = body.encode( 'utf-8' )
            xml.sax.parseString( body, h )
            # noinspection PyUnresolvedReferences
            return rs.LocationConstraint
        else:
            raise bucket.connection.provider.storage_response_error(
                response.status, response.reason, body )


class BucketModification( Operation ):
    """
    An operation that modifies a bucket
    """

    __metaclass__ = abc.ABCMeta

    def __init__( self, dst_url, **kwargs ):
        super( BucketModification, self ).__init__( **kwargs )
        dst_url = urlparse( dst_url )
        if dst_url.scheme == 's3' and dst_url.netloc and dst_url.path.startswith( '/' ):
            self.bucket_name = dst_url.netloc
            assert dst_url.path.startswith( '/' )
            self.key_name = dst_url.path[ 1: ]
            with closing( S3Connection( ) ) as s3:
                headers = self._get_default_headers( )
                bucket = s3.get_bucket( self.bucket_name, headers=headers )
                location = self.get_bucket_location( bucket, headers=headers )
                self.bucket_region = bucket_location_to_region( location )
        else:
            raise UserError( 'Destination URL must be of the form s3://BUCKET/ or s3://BUCKET/KEY' )

    def _get_uploads( self, bucket, limit=max_uploads_per_page, allow_prefix=False ):
        """
        Get all open multipart uploads for the user-specified key
        """
        if limit > max_uploads_per_page:
            raise ValueError( "Limit must not exceed %i" % max_uploads_per_page )
        headers = self._get_default_headers( )
        uploads = bucket.get_all_multipart_uploads( prefix=self.key_name,
                                                    max_uploads=limit,
                                                    headers=headers )
        if len( uploads ) == max_uploads_per_page:
            raise RuntimeError( "Can't handle more than %i uploads" % max_uploads_per_page )

        uploads = [ MultiPartUploadPlus( bucket=bucket,
                                         key_name=upload.key_name,
                                         upload_id=upload.id,
                                         headers=headers )
            for upload in uploads if allow_prefix or upload.key_name == self.key_name ]

        return uploads


# noinspection PyIncorrectDocstring
class MultiPartUploadPlus( MultiPartUpload ):
    """
    There is no built-in way to just get a MultiPartUpload object without its children part
    objects. There is either ListParts or ListMultiPartUploads. We can, however, fake one
    by creating an empty MultiPartObject and fill in the attributes we know are needed for the
    method calls we perform on that object.
    """

    def __init__( self, bucket, key_name, upload_id, parts=None, headers=None ):
        """
        Constructs a multi-part object and optionally pass a list of parts. If the parts are
        supplied, the resulting object can be used to complete the upload via its
        complete_upload() method.

        :type parts: list[Part]
        """
        super( MultiPartUploadPlus, self ).__init__( bucket )
        self.id = upload_id
        self.key_name = key_name
        self._parts = parts
        self._headers = headers

    def __iter__( self ):
        if self._parts is None:
            return super( MultiPartUploadPlus, self ).__iter__( )
        else:
            return iter( self._parts )

    # Work around the fact that Boto's get_all_parts() and __iter__() do not support the headers 
    # keyword argument and the fact that get_all_parts() returns None on error rather than raising 
    # an exception

    def get_all_parts( self, max_parts=None, part_number_marker=None,
                       encoding_type=None ):
        """
        Return the uploaded parts of this MultiPart Upload.  This is
        a lower-level method that requires you to manually page through
        results.  To simplify this process, you can just use the
        object itself as an iterator and it will automatically handle
        all of the paging with S3.
        """
        self._parts = [ ]
        query_args = 'uploadId=%s' % self.id
        if max_parts:
            query_args += '&max-parts=%d' % max_parts
        if part_number_marker:
            query_args += '&part-number-marker=%s' % part_number_marker
        if encoding_type:
            query_args += '&encoding-type=%s' % encoding_type
        conn = self.bucket.connection
        response = conn.make_request( 'GET', self.bucket.name,
                                      self.key_name,
                                      query_args=query_args,
                                      headers=self._headers )
        body = response.read( )
        if response.status == 200:
            h = handler.XmlHandler( self, self )
            xml.sax.parseString( body, h )
            return self._parts
        else:
            raise conn.provider.storage_response_error( response.status, response.reason, body )

    # Work around lack of headers suport in complete_upload()

    def complete_upload( self, headers=None ):
        """
        Complete the MultiPart Upload operation.  This method should
        be called when all parts of the file have been successfully
        uploaded to S3.

        :rtype: :class:`boto.s3.multipart.CompleteMultiPartUpload`
        :returns: An object representing the completed upload.
        """
        return self.bucket.complete_multipart_upload( self.key_name,
                                                      self.id,
                                                      self.to_xml( ),
                                                      headers=headers )

    # Work around lack of headers suport in cancel_upload()

    def cancel_upload( self, headers=None ):
        """
        Cancels a MultiPart Upload operation.  The storage consumed by
        any previously uploaded parts will be freed. However, if any
        part uploads are currently in progress, those part uploads
        might or might not succeed. As a result, it might be necessary
        to abort a given multipart upload multiple times in order to
        completely free all storage consumed by all parts.
        """
        self.bucket.cancel_multipart_upload( self.key_name, self.id, headers=headers )


class Cancel( BucketModification ):
    """
    Cancel an unfinished upload
    """

    def __init__( self, dst_url, allow_prefix, **kwargs ):
        super( Cancel, self ).__init__( dst_url, **kwargs )
        if (self.key_name.endswith( '/' ) or self.key_name == '') and not allow_prefix:
            raise UserError( "Make sure the destination URL does not end in / or pass --prefix if "
                             "you really intend to delete uploads for all objects whose key starts "
                             "with '%s'." % self.key_name )
        self.allow_prefix = allow_prefix

    def run( self ):
        with closing( s3_connect_to_region( self.bucket_region ) ) as s3:
            headers = self._get_default_headers( )
            bucket = s3.get_bucket( self.bucket_name, headers=headers )
            uploads = self._get_uploads( bucket, allow_prefix=self.allow_prefix )
            for upload in uploads:
                upload.cancel_upload( headers=headers )


class Upload( BucketModification ):
    """
    Perform or resume an upload
    """

    def __init__( self, src_url, dst_url,
                  resume=False, force=False, part_size=min_part_size,
                  download_slots=num_cores, upload_slots=num_cores,
                  sse_key=None, src_sse_key=None, **kwargs ):
        super( Upload, self ).__init__( dst_url, **kwargs )
        # Neither dot nor slash occur in a valid URL's scheme part so we can use those to detect
        # a path, even if that path contains a colon. Also anything without a colon can't be a
        # URL and we'll assume it is a path.
        if src_url[ 0 ] in './' or ':' not in src_url:
            src_url = 'file://' + os.path.abspath( src_url )
        parsed_src_url = urlparse( src_url )
        if parsed_src_url.scheme == 'file' and parsed_src_url.netloc not in ('', 'localhost'):
            raise UserError( 'The URL %s is not a valid file:// URL. For absolute paths use '
                             'file:/ABSOLUTE/PATH/TO/FILE, file:///ABSOLUTE/PATH/TO/FILE or just '
                             '/ABSOLUTE/PATH/TO/FILE. For relative paths use '
                             'RELATIVE/PATH/TO/FILE. To refer to a file called FILE in the current '
                             'working directory, use FILE.' )
        src_url = parsed_src_url
        if self.key_name.endswith( '/' ) or self.key_name == '':
            self.key_name += os.path.basename( src_url.path )
        src_url = src_url.geturl( )
        self.url = src_url
        self.part_size = part_size
        self.resume = resume
        self.force = force
        self.download_slots = download_slots
        self.upload_slots = upload_slots
        self.sse_key = sse_key
        self.src_sse_key = src_sse_key

    def run( self ):
        """
        Stream a URL to a key in an S3 bucket using a parallelized multi-part upload.
        """
        global download_slots_semaphore, done_event, error_event

        download_slots_semaphore = multiprocessing.Semaphore( self.download_slots )
        done_event = multiprocessing.Event( )
        error_event = multiprocessing.Event( )

        upload_id, completed_parts = self._prepare_upload( )
        part_nums = itertools.count( )
        num_workers = self.download_slots + self.upload_slots
        workers = multiprocessing.Pool( num_workers, _init_worker )

        # Stagger the workers a bit, to avoid swamping the metadata service
        # See https://github.com/BD2KGenomics/s3am/issues/16 for details
        for _ in range( num_workers ):
            workers.apply_async( functools.partial( time.sleep, random.random( ) + .5 ) )

        def complete_part( ( part_num_, part ) ):
            if part is not None:
                assert part_num_ not in completed_parts
                completed_parts[ part_num_ ] = part

        if self.url.startswith( 's3:' ):
            log.info( 'Copying %s' % self.url )
            url = urlparse( self.url )
            assert url.scheme == 's3'
            assert url.path.startswith( '/' )
            with closing( S3Connection( ) ) as s3:
                headers = self._get_default_headers( )
                src_bucket = s3.get_bucket( url.netloc, headers=headers )
                if self.src_sse_key:
                    self._add_encryption_headers( self.src_sse_key, headers )
                src_key = src_bucket.get_key( url.path[ 1: ], headers=headers )
            worker_func = self._copy_part
            kwargs = dict(
                src_bucket_name=src_bucket.name,
                src_key_name=src_key.name,
                size=src_key.size )
        else:
            log.info( 'Streaming %s' % self.url )
            kwargs = { }
            worker_func = self._stream_part

        try:
            while not done_event.is_set( ):
                if error_event.is_set( ):
                    raise WorkerException( )
                part_num = next( part_nums )
                if part_num in completed_parts:
                    assert self.resume
                    log.info( 'part %i: exists', part_num )
                else:
                    download_slots_semaphore.acquire( )
                    log.info( 'part %i: dispatching', part_num )
                    workers.apply_async( func=worker_func,
                                         args=[ upload_id, part_num ],
                                         kwds=kwargs,
                                         callback=complete_part )
            workers.close( )
            workers.join( )
            if error_event.is_set( ):
                raise WorkerException( )
            self._sanity_check( completed_parts )
            with closing( s3_connect_to_region( self.bucket_region ) ) as s3:
                headers = self._get_default_headers( )
                bucket = s3.get_bucket( self.bucket_name, headers=headers )
                upload = MultiPartUploadPlus( bucket=bucket,
                                              key_name=self.key_name,
                                              upload_id=upload_id,
                                              parts=completed_parts.values( ) )
                upload.complete_upload( headers=headers )
            log.info( 'Completed %s' % self.url )
        except WorkerException:
            workers.close( )
            workers.join( )
            raise
        except (Exception, KeyboardInterrupt):
            workers.close( )
            workers.terminate( )
            raise

    def _prepare_upload( self ):
        """
        Prepare a new multipart upload or resume a previously interrupted one. Returns the upload
        ID and a dictionary mapping the 0-based index of a part to its size.
        """
        with closing( s3_connect_to_region( self.bucket_region ) ) as s3:
            headers = self._get_default_headers( )
            bucket = s3.get_bucket( self.bucket_name, headers=headers )
            uploads = self._get_uploads( bucket )
            while True:
                if len( uploads ) == 0:
                    headers = self._get_default_headers( )
                    if self.sse_key:
                        self._add_encryption_headers( self.sse_key, headers )
                    # Necessary when uploader and bucket owner are differnt AWS accounts. Without
                    # this, only an ACL for the uploader will be created. With this header,
                    # an ACL for the uploader and one for the bucket owner will be created.
                    headers['x-amz-acl'] = 'bucket-owner-full-control'
                    upload_id = bucket.initiate_multipart_upload( key_name=self.key_name,
                                                                  headers=headers ).id
                    return upload_id, { }
                elif self.resume:
                    if len( uploads ) == 1:
                        upload = uploads[ 0 ]
                        completed_parts = { part.part_number - 1: part for part in upload }
                        # If there is an upload but no parts we can use whatever part size we want,
                        # otherwise we need to ensure that we use the same part size as before.
                        if len( completed_parts ) > 0:
                            previous_part_size = self._guess_part_size( completed_parts )
                            if self.part_size != previous_part_size:
                                raise UserError(
                                    "Transfer failed. The part size appears to have changed from "
                                    "%i to %i. Either resume the upload with the previous part "
                                    "size or cancel it before using the new part size." % (
                                        self.part_size, previous_part_size) )
                        return upload.id, completed_parts
                    else:
                        raise RuntimeError(
                            "Transfer failed. There are multiple unfinished uploads, so there is "
                            "ambiguity as to which one of them should be resumed. Consider using "
                            "'{me} cancel s3://{bucket_name}/{key_name}' to delete all of them "
                            "before trying the transfer again. Note that unfinished uploads incur "
                            "storage fees.".format( me=me, **vars( self ) ) )
                elif self.force:
                    log.warn( 'Cancelling all unfinished uploads before proceeding.' )
                    for upload in uploads:
                        bucket.cancel_multipart_upload( upload.key_name,
                                                        upload.id,
                                                        headers=self._get_default_headers( ) )
                    uploads = [ ]
                else:
                    if len( uploads ) == 1:
                        raise UserError(
                            "Transfer failed. There is an unfinished upload. To resume that "
                            "upload, run {me} again with --resume. To cancel it, use '{me} cancel "
                            "s3://{bucket_name}/{key_name}'. Note that unfinished uploads incur "
                            "storage fees.".format( me=me, **vars( self ) ) )
                    else:
                        raise RuntimeError(
                            "Transfer failed. Detected unfinished multipart uploads. Consider "
                            "using '{me} cancel s3://{bucket_name}/{key_name}' to delete all of "
                            "them before trying the transfer again. Note that pending uploads "
                            "incur storage fees.".format( me=me, **vars( self ) ) )

    def _guess_part_size( self, completed_parts ):
        size_groups = self._part_size_histogram( completed_parts )
        assert len( size_groups ) > 0
        # We can't handle more than two different sizes (first term) and if we have two different
        # sizes, the smaller one should only occur once (second term).
        if len( size_groups ) > 2 \
            or len( size_groups ) == 2 and size_groups[ 1 ][ 1 ] != 1:
            raise RuntimeError(
                "Can't reliably determine previously used part size for this upload. "
                "You should probably cancel it and start over." )
        return size_groups[ 0 ][ 0 ]

    @contextmanager
    def _propagate_worker_exception( self ):
        """
        A context manager that sets the error_event if any exception occurs in the body. It can
        be safely assumed that this context manager does not cause any exceptions before it
        yields to the enclosed block.
        """
        try:
            yield
        except BailoutException:
            raise
        except BaseException:
            error_event.set( )
            log.error( traceback.format_exc( ) )
            raise

    def _stream_part( self, upload_id, part_num ):
        """
        Download o part from the source URL, buffer it in memory and then upload it to S3.
        """
        with self._propagate_worker_exception( ):
            try:
                log.info( 'part %i: downloading', part_num )
                buf = self._download_part( part_num )
            finally:
                download_slots_semaphore.release( )
            download_size = buf.tell( )
            self._log_progress( part_num, 'downloaded', download_size )
            if download_size > self.part_size:
                assert False
            elif download_size < self.part_size:
                done_event.set( )
            else:
                pass
            if error_event.is_set( ): raise BailoutException( )
            if download_size > 0 or part_num == 0:
                log.info( 'part %i: uploading', part_num )
                buf.seek( 0 )
                part = self._upload_part( upload_id, part_num, buf )
                upload_size = buf.tell( )
                assert download_size == upload_size == part.size
                self._log_progress( part_num, 'uploaded', upload_size )
            else:
                part = None
            return part_num, part

    def _log_progress( self, part_num, task, download_size ):
        log.info( 'part %i: %s %sB (%i bytes)',
                  part_num, task, bytes2human( download_size ), download_size )

    def _download_part( self, part_num ):
        """
        Download a part from the source URL. Returns a BytesIO buffer. The buffer's tell() method
        will return the size of the downloaded part, which may be less than the requested part
        size if the part is the last one for the URL.
        """
        buf = BytesIO( )
        with closing( pycurl.Curl( ) ) as c:
            c.setopt( c.URL, self.url )
            c.setopt( c.WRITEDATA, buf )
            c.setopt( c.FAILONERROR, 1 )
            start, end = self._get_part_range( part_num )
            c.setopt( c.RANGE, "%i-%i" % (start, end - 1) )
            try:
                c.perform( )
            except pycurl.error as e:
                error_code, message = e
                if error_code == c.E_BAD_DOWNLOAD_RESUME:  # bad range for FTP
                    pass
                elif error_code == c.E_HTTP_RETURNED_ERROR:
                    # http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html#sec10.4.17
                    if c.getinfo( c.RESPONSE_CODE ) == 416:
                        pass
                    else:
                        raise
                else:
                    raise
                buf.truncate( 0 )
                buf.seek( 0 )
        return buf

    def _get_part_range( self, part_num ):
        """
        Returns a tuple (start,end) with start being the offset of the first byte to copy and end
        being the offset of the byte after the last byte to copy.

        Invariant:  end - start == self.part_size
        """
        start = part_num * self.part_size
        end = start + self.part_size
        return start, end

    def _upload_part( self, upload_id, part_num, buf ):
        """
        Upload a part to S3. The given buffer's tell() is assumed to point at the first byte to
        be uploaded. When this method returns, the tell() method points at the end of the buffer.

        :return: the part object representing the uploaded part
        :rtype: Part
        """
        with closing( s3_connect_to_region( self.bucket_region ) ) as s3:
            headers = self._get_default_headers( )
            bucket = s3.get_bucket( self.bucket_name, headers=headers )
            upload = MultiPartUploadPlus( bucket=bucket,
                                          key_name=self.key_name,
                                          upload_id=upload_id,
                                          headers=headers )
            if self.sse_key:
                self._add_encryption_headers( self.sse_key, headers )
            key = upload.upload_part_from_file( buf, part_num + 1, headers=headers )
            return self._part_for_key( bucket, part_num, key )

    def _copy_part( self, upload_id, part_num, src_bucket_name, src_key_name, size ):
        with self._propagate_worker_exception( ):
            try:
                if error_event.is_set( ): raise BailoutException( )
                log.info( 'part %i: copying', part_num )
                start, end = self._get_part_range( part_num )
                end = min( end, size )
                part_size = end - start
                if part_size > 0 or part_num == 0:
                    url = urlparse( self.url )
                    assert url.scheme == 's3'
                    assert url.path.startswith( '/' )
                    with closing( s3_connect_to_region( self.bucket_region ) ) as s3:
                        headers = self._get_default_headers( )
                        bucket = s3.get_bucket( self.bucket_name, headers=headers )
                        upload = MultiPartUploadPlus( bucket=bucket,
                                                      key_name=self.key_name,
                                                      upload_id=upload_id,
                                                      headers=headers )
                        if self.sse_key:
                            self._add_encryption_headers( self.sse_key, headers )
                        if part_size == 0:
                            # Since copy_part_from_key doesn't allow empty ranges, we handle that
                            # case by uploading an empty part.
                            assert part_num == 0
                            # noinspection PyTypeChecker
                            key = upload.upload_part_from_file(
                                StringIO( ), part_num + 1,
                                headers=headers )
                        else:
                            if self.src_sse_key:
                                self._add_encryption_headers( self.src_sse_key, headers,
                                                              for_copy=True )
                            key = upload.copy_part_from_key(
                                src_bucket_name=src_bucket_name,
                                src_key_name=src_key_name,
                                part_num=part_num + 1,
                                start=start,
                                end=end - 1,
                                headers=headers )
                            # somehow copy_part_from_key doesn't set the key size
                            key.size = part_size
                    assert key.size == part_size
                    self._log_progress( part_num, 'copied', part_size )
                    return part_num, self._part_for_key( bucket, part_num, key )
                else:
                    done_event.set( )
                    return part_num, None
            finally:
                download_slots_semaphore.release( )

    def _part_for_key( self, bucket, part_num, key ):
        """
        Most of boto's multipart-related methods return a Key object rather than a part object.
        This method converts the former to the latter. Ultimately, we need a MultiPartUpload with
        a list of Part objects in order to complete a multi-part upload.

        :rtype : Part
        """
        part = Part( )
        part.bucket = bucket
        part.part_number = part_num + 1
        part.size = key.size
        part.etag = key.etag
        part.last_modified = key.last_modified
        return part

    def _sanity_check( self, completed_parts ):
        """
        Verify that all parts are present and valid.
        """

        # Check uniqueness property (no duplicate part_nums)
        assert isinstance( completed_parts, dict )

        # At least one part
        assert len( completed_parts ) > 0

        # Check completeness (no missing part nums)
        assert max( completed_parts ) >= len( completed_parts ) - 1

        # The parts should belong to exaclty one or two distinct size groups.
        groups = self._part_size_histogram( completed_parts )

        def sentinel( part_size, num_parts ):
            # S3 needs at least one part per upload so for empty files we need a sentinel part
            return part_size == 0 and num_parts == 1

        def one_or_more_full_parts( part_size, num_parts ):
            return part_size == self.part_size and num_parts > 0

        def one_last_part( part_size, num_parts ):
            return 0 < part_size < self.part_size and num_parts == 1

        def xor( *bools ):
            return bools.count( True ) == 1

        if len( groups ) == 1:
            assert xor( sentinel( *groups[ 0 ] ),
                        one_or_more_full_parts( *groups[ 0 ] ),
                        one_last_part( *groups[ 0 ] ) )
        elif len( groups ) == 2:
            assert one_or_more_full_parts( *groups[ 0 ] ) and one_last_part( *groups[ 1 ] )
        else:
            assert False

    def _part_size_histogram( self, completed_parts ):
        """
        Group input parts by size and return the length of each group.

        :param completed_parts: a dictionary mapping part numbers to part sizes

        :type completed_parts: dict[int,boto.s3.key.Key]

        :return: A list of (part_size, num_parts) tuples where part_size is the size of a part and
        num_parts the number of parts of that size in the input. The returned list is sorted by
        descending part size. Note that there can't be any empty groups in the result, so the second
        member of each tuple is guaranteed to be greater than zero.
        """

        by_part_size = itemgetter( 1 )

        # Convert to list of ( part_num, part_size ) pairs, sorted by descending size
        completed_parts = [ (part_num, part.size) for part_num, part in
            completed_parts.iteritems( ) ]
        completed_parts.sort( key=by_part_size, reverse=True )

        # Count # of elements in iterator
        def ilen( it ):
            return sum( 1 for _ in it )

        return [ (part_size, ilen( group )) for part_size, group in
            itertools.groupby( completed_parts, by_part_size ) ]


class BailoutException( RuntimeError ):
    pass


def _init_worker( ):
    """
    Hacks around weirdness with the multiprocessing module and Ctrl-C
    """
    signal.signal( signal.SIGINT, signal.SIG_IGN )


# Allow instance methods to be used with multiprocessing
#
# Stolen from https://gist.github.com/fiatmoney/1086393

def _pickle_method( method ):
    func_name = method.im_func.__name__
    obj = method.im_self
    cls = method.im_class
    if func_name.startswith( '__' ) and not func_name.endswith( '__' ):  # deal with mangled names
        cls_name = cls.__name__.lstrip( '_' )
        func_name += '_' + cls_name
    return _unpickle_method, (func_name, obj, cls)


def _unpickle_method( func_name, obj, cls ):
    for cls in cls.__mro__:
        try:
            func = cls.__dict__[ func_name ]
        except KeyError:
            pass
        else:
            return func.__get__( obj, cls )
    assert False


import copy_reg
import types

copy_reg.pickle( types.MethodType, _pickle_method, _unpickle_method )


class Verify( Operation ):
    """
    Compute a checksum of the content at a given URL.
    """

    def __init__( self, url, checksum, sse_key, part_size, **kwargs ):
        super( Verify, self ).__init__( **kwargs )
        url = urlparse( url )
        if url.scheme != 's3':
            raise NotImplementedError(
                "Only 's3' URLs are currently supported, not '%s." % url.scheme )
        if not url.netloc or not url.path.startswith( '/' ):
            raise UserError( 'An S3 URL must be of the form s3:/BUCKET/ or s3://BUCKET/KEY' )
        self.url = url
        self.checksum = checksum
        self.sse_key = sse_key
        self.part_size = part_size

    def run( self ):
        assert self.url.scheme == 's3' and self.url.netloc and self.url.path.startswith( '/' )
        with closing( S3Connection( ) ) as s3:
            headers = self._get_default_headers( )
            bucket = s3.get_bucket( self.url.netloc, headers=headers )
            key = bucket.get_key( self.url.path[ 1: ], headers=headers )
            if self.sse_key:
                self._add_encryption_headers( self.sse_key, headers )
            start, size = 0, key.size
            while start < size:
                end = min( start + self.part_size, size )
                key.open_read( headers=dict( headers,
                                             Range='bytes=%i-%i' % (start, end - 1) ) )
                try:
                    while True:
                        buf = key.read( verify_buffer_size )
                        if not buf: break
                        self.checksum.update( buf )
                finally:
                    key.close( )
                start = end
        return self.checksum.hexdigest( )



# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import hashlib
import logging
import sys
import inspect
import argparse

from s3am import UserError
from s3am.humanize import human2bytes
from s3am.operations import (min_part_size,
                             max_part_size,
                             max_parts_per_upload,
                             Upload,
                             Cancel,
                             Verify)


def try_main( args=sys.argv[ 1: ] ):
    """
    Main entry point. Neither thread-safe nor reentrant but can be called repeatedly.
    """
    try:
        main( args )
    except UserError as e:
        sys.stderr.write( "error: %s\n" % e.message )
        sys.exit( 2 )


def main( args ):
    o = parse_args( args )
    if o.verbose:
        logging.getLogger( ).setLevel( logging.INFO )
    if o.debug:
        logging.getLogger( ).setLevel( logging.DEBUG )
    kwargs = dict( requester_pays=o.requester_pays )
    if o.mode == 'upload':
        operation = Upload(
            src_url=o.src_url,
            dst_url=o.dst_url,
            resume=o.resume,
            force=o.force,
            part_size=o.part_size,
            download_slots=o.download_slots,
            upload_slots=o.upload_slots,
            sse_key=o.sse_key or o.sse_key_file or o.sse_key_base64,
            src_sse_key=o.src_sse_key or o.src_sse_key_file or o.src_sse_key_base64,
            **kwargs )
    elif o.mode == 'cancel':
        operation = Cancel(
            dst_url=o.dst_url,
            allow_prefix=o.allow_prefix,
            **kwargs )
    elif o.mode == 'verify':
        try:
            checksum = hashlib.new( o.checksum )
        except ValueError:
            raise UserError( "Checksum algorithm '%s' does not exist" % o.checksum )
        operation = Verify(
            url=o.url,
            checksum=checksum,
            sse_key=o.sse_key or o.sse_key_file or o.sse_key_base64,
            part_size=o.part_size,
            **kwargs )
    else:
        assert False
    result = operation.run( )
    if result is not None:
        print result
    return result


def default_args( function ):
    """
    Determine the default values for keyword arguments in the given function. Returns a
    dictionary mapping argument names to default values.

    >>> def f(foo=123): pass
    >>> default_args( f )['foo']
    123
    """
    spec = inspect.getargspec( function )
    return dict( zip( spec.args[ -len( spec.defaults ): ], spec.defaults ) )


def parse_args( args ):
    """
    Parse command line arguments and set global option variable with parse result
    """
    p = argparse.ArgumentParser( add_help=False,
                                 description="Stream content from HTTP or FTP servers to S3.",
                                 formatter_class=argparse.ArgumentDefaultsHelpFormatter )

    def add_common_arguments( sp ):
        sp.add_argument( '--verbose', action='store_true',
                         help="Print informational log messages." )
        sp.add_argument( '--debug', action='store_true',
                         help="Print debug log messages. WARNING: This will leak encryption keys!" )
        sp.add_argument( '--requester-pays', action='store_true',
                         help="Agree to be charged for requests against buckets not owned by you. "
                              "For details refer to http://docs.aws.amazon.com/AmazonS3/latest/dev"
                              "/RequesterPaysBuckets.html" )

    p.add_argument( '--help', action=ArgParseOverallHelpAction, help="Show this help and exit." )

    sps = p.add_subparsers( dest='mode' )

    upload_sp = sps.add_parser( 'upload', add_help=False, help="Perform an upload.",
                                description="Download the contents of the given URL and upload it "
                                            "to the specified key and bucket in S3 using multiple "
                                            "processes in parallel.",
                                formatter_class=argparse.ArgumentDefaultsHelpFormatter )

    gr = upload_sp.add_mutually_exclusive_group( )
    gr.add_argument( '--resume', action='store_true',
                     help="Resume a previously interrupted and therefore unfinished upload "
                          "for the given object if such an upload exists. When resuming an "
                          "upload, already uploaded parts will be skipped. Be advised that "
                          "unfinished uploads are more or less hidden objects that "
                          "nevertheless incur storage fees, just like regular objects. Use "
                          "'s3am cancel' to remove unfinished uploads for a given object." )

    gr.add_argument( '--force', action='store_true',
                     help="Delete all previously interrupted and therefore unfinished "
                          "uploads for the given object before beginning a new upload. "
                          "Without this flag, s3am will err on the side of caution and "
                          "exit with an error if it detects unfinished uploads." )

    defaults = default_args( Upload.__init__ )

    upload_sp.add_argument( '--download-slots', type=int, metavar='NUM',
                            default=defaults[ 'download_slots' ],
                            help="The number of processes that will concurrently upload to S3." )
    upload_sp.add_argument( '--upload-slots', type=int, metavar='NUM',
                            default=defaults[ 'download_slots' ],
                            help="The number of processes that will concurrently download from "
                                 "the SRC_URL." )

    def parse_part_size( s ):
        i = human2bytes( s )
        if i < min_part_size:
            raise argparse.ArgumentTypeError( "Part size must be at least %i" % min_part_size )
        if i > max_part_size:
            raise argparse.ArgumentTypeError( "Part size must not exceed %i" % max_part_size )
        return i

    upload_sp.add_argument( '--part-size', metavar='NUM',
                            default=defaults[ 'part_size' ], type=parse_part_size,
                            help="The number of bytes in each part. This parameter must be at "
                                 "least {min} and no more than {max}. The default is {min}. Note "
                                 "that S3 allows no more than {max_parts} per upload and this "
                                 "program does not currently ensure that this parameter is large "
                                 "enough to stream the SRC_URL's content in its entirety using "
                                 "those {max_parts} parts.".format( min=min_part_size,
                                                                    max=max_part_size,
                                                                    max_parts=max_parts_per_upload ) )

    def parse_sse_key( s ):
        if len( s ) != 32:
            raise argparse.ArgumentTypeError( "SSE-C key must be exactly 32 bytes long" )
        return s

    def parse_sse_key_file( s ):
        with open( s ) as f:
            return parse_sse_key( f.read( ) )

    def parse_sse_key_base64( s ):
        return parse_sse_key( base64.b64decode( s ) )

    def add_sse_opts( sp, helps ):
        for prefix, sse_help in helps.iteritems( ):
            sse_key_gr = sp.add_mutually_exclusive_group( )
            sse_key_gr.add_argument( prefix, metavar='KEY', type=parse_sse_key,
                                     help="The %s. If the key starts with a - (dash) character, "
                                          "the --sse-key=... form of this option must be used." %
                                          sse_help )
            sse_key_gr.add_argument( prefix + '-file', metavar='PATH', type=parse_sse_key_file,
                                     help="The path to a file containing the %s." % sse_help )
            sse_key_gr.add_argument( prefix + '-base64', metavar='KEY', type=parse_sse_key_base64,
                                     help="The base64 encoding of the %s" % sse_help )

    def sse_help( purpose='' ):
        return ("binary 32-byte key to use for " + purpose + "server-side encryption with "
                                                             "customer-provided keys (SSE-C).")

    add_sse_opts( upload_sp, {
        '--sse-key': sse_help( ) + " " +
                     "The given key will be used to encrypt the uploaded content at rest in S3. "
                     "Subsequent downloads of the object will require the same key",
        '--src-sse-key': sse_help( purpose="copying an S3 object that uses " ) + " " +
                         "This option is only applicable if SRC_URL refers starts with s3://." } )

    add_common_arguments( upload_sp )

    upload_sp.add_argument( 'src_url', metavar='SRC_URL',
                            help="The location to read from. S3AM currently support 's3:', "
                                 "'file:', 'http:' and 'ftp:' URLs. Depending on the local "
                                 "libcurl installation, additional URL schemes, like 'https:' may "
                                 "be supported. An S3 URL has the form 's3://BUCKET/KEY'. The URL "
                                 "of a local file has the form 'file:/PATH', "
                                 "'file://localhost/PATH' or 'file:///PATH' where PATH is the "
                                 "absolute path to a file without the leading slash. Instead of a "
                                 "'file:' URL pointing to a local file, just the absolute path to "
                                 "that local file may be specified. Likewise, a relative path to "
                                 "a local file may be specified, provided that it starts with "
                                 "'./' or contains no ':' characters.")
    upload_sp.add_argument( 'dst_url', metavar='DST_URL',
                            help="The location of the S3 object to write to. Must be of the form "
                                 "s3://BUCKET/KEY. If DST_URL ends in a slash, the last path "
                                 "component from SRC_URL will be appended to DST_URL." )

    cancel_sp = sps.add_parser( 'cancel', add_help=False, help="Cancel unfinished uploads.",
                                description="Cancel multipart uploads that were not completed.",
                                formatter_class=argparse.ArgumentDefaultsHelpFormatter )

    add_common_arguments( cancel_sp )

    cancel_sp.add_argument( 'dst_url', metavar='URL',
                            help="The S3 URL for which to delete pending uploads. Must be of the "
                                 "form s3://BUCKET/KEY. URL must not end in a slash unless "
                                 "--prefix is passed." )

    cancel_sp.add_argument( '--prefix', dest='allow_prefix', action='store_true',
                            help="Treat URL as a prefix, i.e. cancel pending uploads for all "
                                 "objects whose URL starts with the given URL. By default only "
                                 "the object whose URL is an exact match will be deleted. In "
                                 "order to delete all uploads for all keys in a bucket, "
                                 "use --prefix and a URL of the form s3://BUCKET/." )

    verify_sp = sps.add_parser( 'verify', add_help=False, help="Verify the contents of a URL.",
                                description="Compute a checksum of an object at a given URL.",
                                formatter_class=argparse.ArgumentDefaultsHelpFormatter )

    add_common_arguments( verify_sp )

    verify_sp.add_argument( 'url', metavar='URL',
                            help="The location of the S3 object to verify. Must be of the form "
                                 "s3://BUCKET/KEY.")

    # algorithms_available was introduced in 2.7.9
    algorithms = getattr( hashlib, 'algorithms_available', None ) or hashlib.algorithms
    verify_sp.add_argument( '--checksum', metavar='TYPE',
                            choices=algorithms, default='md5',
                            help="The checksum algorithm to use for verification. Valid choices "
                                 "are %s." % ', '.join( algorithms ) )

    add_sse_opts( verify_sp, {
        '--sse-key': "binary 32-byte key to use for verifying an S3 object that is encrypted with "
                     "server-side encryption using customer-provided keys (SSE-C)." } )

    def parse_verify_part_size( s ):
        i = human2bytes( s )
        if i < 1:
            raise argparse.ArgumentTypeError( "Part size must be at least 1" )
        return i

    verify_sp.add_argument( '--part-size', metavar='NUM',
                            default=defaults[ 'part_size' ], type=parse_verify_part_size,
                            help="The number of bytes in each part to verify. Verification is "
                                 "broken into parts for increased robustness." )
    return p.parse_args( args )


# noinspection PyProtectedMember
class ArgParseOverallHelpAction( argparse._HelpAction ):
    def __call__( self, parser, namespace, values, option_string=None ):
        parser.print_help( )

        # retrieve subparsers from parser
        subparsers_actions = [
            action for action in parser._actions
            if isinstance( action, argparse._SubParsersAction ) ]
        # there will probably only be one subparser_action,
        # but better save than sorry
        for subparsers_action in subparsers_actions:
            # get all subparsers and print help
            for choice, subparser in subparsers_action.choices.items( ):
                sys.stderr.write( "\n\n\n".format( choice ) )
                sys.stderr.write( subparser.format_help( ) )

        parser.exit( )



# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import sys

me = os.path.basename( sys.argv[ 0 ] )

log = logging.getLogger( __name__ )

logging.basicConfig( level=logging.WARN,
                     format="%(asctime)-15s %(module)s(%(process)d) %(message)s" )


class UserError( Exception ):
    """
    An exception that doesn't cause a stack trace to be printed.
    """
    pass


class WorkerException( Exception ):
    """
    An exception where we let other workers finish
    """
    pass



# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

from pyftpdlib.authorizers import DummyAuthorizer
from pyftpdlib.handlers import FTPHandler
from pyftpdlib.servers import FTPServer


class FTPd( threading.Thread ):
    def __init__( self, root_dir, address=None, timeout=0.001, dtp_handler=None):
        threading.Thread.__init__( self )
        self.__flag = threading.Event( )
        self.__timeout = timeout
        authorizer = DummyAuthorizer( )
        authorizer.add_anonymous( root_dir )
        handler = FTPHandler
        handler.authorizer = authorizer
        if dtp_handler is not None:
            handler.dtp_handler = dtp_handler
        self.server = FTPServer( address, handler )

    def start( self ):
        self.__flag.clear( )
        threading.Thread.start( self )
        self.__flag.wait( )

    def run( self ):
        self.__flag.set( )
        while self.__flag.is_set( ):
            self.server.serve_forever( timeout=self.__timeout, blocking=False )
        self.server.close_all( )
        self.server.close()

    def stop( self ):
        self.__flag.clear( )
        self.join( )



# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import logging
import socket
import time
import unittest
from contextlib import closing
from tempfile import mkdtemp
from threading import Lock

import FTPd
import boto.exception
import boto.s3
import os
import pyftpdlib.handlers
import s3am
import s3am.boto_utils
import s3am.cli
import s3am.operations

# The dot in the domain name makes sure that boto.work_around_dots_in_bucket_names() is covered
from bd2k.util.iterables import concat

test_bucket_name_prefix = 's3am-unit-tests.foo'
test_bucket_region = 'us-west-1'
copy_bucket_region = 'us-east-1' # using us-east-1 so we get exposed to its quirks

host = "127.0.0.1"
port = 21212
part_size = s3am.operations.min_part_size
two_and_a_half_parts = int( part_size * 2.5 )
two_parts = 10 * 1024 * 1024
test_sizes = [ 0, 1, part_size - 1, part_size, part_size + 1, two_parts, two_and_a_half_parts ]

verbose = ('--verbose',)  # ('--debug',)

num_slots = 4  # how many download and upload slots to use

log = logging.getLogger( __name__ )

slots = ('--download-slots', str( num_slots ), '--upload-slots', str( num_slots ))
one_slot = ('--download-slots', '1', '--upload-slots', '0')


def md5( contents ):
    return hashlib.md5( contents ).hexdigest( )


class TestFile( object ):
    def __init__( self, ftp_root, size ):
        self.ftp_root = ftp_root
        self.size = size
        buf = bytearray( os.urandom( size ) )
        self.md5 = md5( buf )
        with open( self.path, 'w' ) as f:
            f.write( buf )

    @property
    def name( self ):
        return 'test-%i.bin' % self.size

    @property
    def path( self ):
        return os.path.join( self.ftp_root, self.name )


class OperationsTests( unittest.TestCase ):
    @classmethod
    def setUpClass( cls ):
        super( OperationsTests, cls ).setUpClass( )
        s3am.boto_utils.work_around_dots_in_bucket_names( )

    def setUp( self ):
        super( OperationsTests, self ).setUp( )
        self.netloc = '%s:%s' % (host, port)
        self.src_url = 'ftp://%s/' % self.netloc
        self.s3 = s3am.boto_utils.s3_connect_to_region( test_bucket_region )
        self.test_bucket_name = '%s-%i' % (test_bucket_name_prefix, int( time.time( ) ))
        test_bucket_location = s3am.boto_utils.region_to_bucket_location( test_bucket_region )
        self.bucket = self.s3.create_bucket( self.test_bucket_name, location=test_bucket_location )
        self._clean_bucket( self.bucket )
        self.ftp_root = mkdtemp( prefix=__name__ )
        self.test_files = { size: TestFile( self.ftp_root, size ) for size in test_sizes }
        self.ftpd = FTPd.FTPd( self.ftp_root, address=(host, port), dtp_handler=UnreliableHandler )
        logging.getLogger( 'pyftpdlib' ).setLevel( logging.WARN )
        self.ftpd.start( )

    def _clean_bucket( self, bucket ):
        for upload in bucket.list_multipart_uploads( ):
            upload.cancel_upload( )
        for key in bucket.list( ):
            key.delete( )

    def tearDown( self ):
        self.ftpd.stop( )
        self._clean_bucket( self.bucket )
        self.bucket.delete( )
        self.s3.close( )
        for test_file in self.test_files.itervalues( ):
            os.unlink( test_file.path )
        os.rmdir( self.ftp_root )

    def _assert_key( self, test_file, sse_key=None ):
        headers = { }
        if sse_key is not None:
            s3am.operations.Upload._add_encryption_headers( sse_key, headers )
        key = self.bucket.get_key( test_file.name, headers=headers )
        self.assertEquals( key.size, test_file.size )
        self.assertEquals( md5( key.get_contents_as_string( headers=headers ) ), test_file.md5 )

    def test_file_urls(self):
        test_file = self.test_files[1]
        for url_prefix in 'file:', 'file://', 'file://localhost':
            s3am.cli.main( concat(
                'upload', verbose, slots,
                url_prefix + test_file.path, self.dst_url( ) ) )
            self._assert_key( test_file )

    def test_invalid_file_urls(self):
        test_file = self.test_files[1]
        for url_prefix in ('file:/',):
            self.assertRaises( s3am.UserError, s3am.cli.main, concat(
                'upload', verbose, slots,
                url_prefix + test_file.path, self.dst_url( ) ) )

    def test_file_path(self):
        test_file = self.test_files[1]
        for path in test_file.path, os.path.relpath(test_file.path):
            s3am.cli.main( concat(
                'upload', verbose, slots,
                path, self.dst_url( ) ) )
            self._assert_key( test_file )

    def test_upload( self ):
        for test_file in self.test_files.itervalues( ):
            s3am.cli.main( concat(
                'upload', verbose, slots,
                self.src_url + test_file.name, self.dst_url( ) ) )
            self._assert_key( test_file )

    def dst_url( self, bucket_name=None, file_name=None ):
        return 's3://%s/%s' % (bucket_name or self.test_bucket_name, file_name or '')

    def test_encryption( self ):
        test_file = self.test_files[ two_and_a_half_parts ]
        src_url = self.src_url + test_file.name
        sse_key = '-0123456789012345678901234567890'
        s3am.cli.main( concat(
            'upload', verbose, slots, src_url, self.dst_url( ),
            '--sse-key=' + sse_key ) )
        self._assert_key( test_file, sse_key=sse_key )
        # Ensure that we can't actually retrieve the object without specifying an encryption key
        try:
            self._assert_key( test_file )
        except boto.exception.S3ResponseError as e:
            self.assertEquals( e.status, 400 )
        else:
            self.fail( )

    def test_resume( self ):
        test_file = self.test_files[ two_and_a_half_parts ]
        src_url = self.src_url + test_file.name

        # Run with a simulated download failure
        UnreliableHandler.setup_for_failure_at( int( 0.75 * test_file.size ) )
        try:
            s3am.cli.main( concat( 'upload', verbose, one_slot, src_url, self.dst_url( ) ) )
        except s3am.WorkerException:
            pass
        else:
            self.fail( )

        # Retrying without --resume should fail
        try:
            s3am.cli.main( concat( 'upload', verbose, slots, src_url, self.dst_url( ) ) )
        except s3am.UserError as e:
            self.assertIn( "unfinished upload", e.message )
        else:
            self.fail( )

        # Retrying with --resume and different part size should fail
        try:
            s3am.cli.main( concat(
                'upload', verbose, slots, src_url, self.dst_url( ),
                '--resume', '--part-size', str( 2 * part_size ) ) )
        except s3am.UserError as e:
            self.assertIn( "part size appears to have changed", e.message )
        else:
            self.fail( )

        # Retry
        s3am.cli.main( concat( 'upload', verbose, slots, src_url, self.dst_url( ), '--resume' ) )

        # FIMXE: We should assert that the resume skips existing parts

        self._assert_key( test_file )

    def test_force( self ):
        test_file = self.test_files[ two_and_a_half_parts ]
        src_url = self.src_url + test_file.name

        # Run with a simulated download failure
        UnreliableHandler.setup_for_failure_at( int( 0.75 * test_file.size ) )
        try:
            s3am.cli.main( concat( 'upload', verbose, one_slot, src_url, self.dst_url( ) ) )
        except s3am.WorkerException:
            pass
        else:
            self.fail( )

        # Retrying without --resume should fail
        try:
            s3am.cli.main( concat( 'upload', verbose, slots, src_url, self.dst_url( ) ) )
        except s3am.UserError as e:
            self.assertIn( "unfinished upload", e.message )
        else:
            self.fail( )

        # Retrying with --force should suceed. We use a different part size to ensure that
        # transfer is indeed started from scratch, not resumed.
        s3am.cli.main( concat(
            'upload', verbose, slots, src_url, self.dst_url( ),
            '--force', '--part-size', str( 2 * part_size ) ) )

    def test_cancel( self ):
        test_file = self.test_files[ two_and_a_half_parts ]
        src_url = self.src_url + test_file.name

        # Run with a simulated download failure
        UnreliableHandler.setup_for_failure_at( int( 0.75 * test_file.size ) )
        try:
            s3am.cli.main( concat( 'upload', verbose, one_slot, src_url, self.dst_url( ) ) )
        except s3am.WorkerException:
            pass
        else:
            self.fail( )

        # A retry without --resume should fail.
        try:
            s3am.cli.main( concat( 'upload', verbose, slots, src_url, self.dst_url( ) ) )
        except s3am.UserError as e:
            self.assertIn( "unfinished upload", e.message )
        else:
            self.fail( )

        # Cancel
        s3am.cli.main( concat( 'cancel', verbose, self.dst_url( file_name=test_file.name ) ) )

        # Retry, should succeed
        s3am.cli.main( concat( 'upload', verbose, slots, src_url, self.dst_url( ) ) )

    def test_copy( self ):
        # setup already created the destination bucket
        dst_bucket_name = self.test_bucket_name
        src_bucket_name = dst_bucket_name + '-src'
        with closing( s3am.boto_utils.s3_connect_to_region( copy_bucket_region ) ) as s3:
            src_location = s3am.boto_utils.region_to_bucket_location( copy_bucket_region )
            src_bucket = s3.create_bucket( src_bucket_name, location=src_location )
            try:
                self._clean_bucket( src_bucket )
                for test_file in self.test_files.itervalues( ):
                    src_url = self.src_url + test_file.name
                    src_sse_key = '-0123456789012345678901234567890'
                    dst_sse_key = 'skdjfh9q4rusidfjs9fjsdr9vkfdh833'
                    dst_url = self.dst_url( src_bucket_name, test_file.name )
                    s3am.cli.main( concat(
                        'upload', verbose, slots, src_url, dst_url,
                        '--sse-key=' + src_sse_key ) )
                    src_url = dst_url
                    dst_url = self.dst_url( )
                    s3am.cli.main( concat(
                        'upload', verbose, slots, src_url, dst_url,
                        '--src-sse-key=' + src_sse_key,
                        '--sse-key=' + dst_sse_key ) )
                    self._assert_key( test_file, dst_sse_key )
            finally:
                self._clean_bucket( src_bucket )
                src_bucket.delete( )

    def test_verify( self ):
        for test_file in self.test_files.itervalues( ):
            s3am.cli.main( concat(
                'upload', verbose, slots,
                self.src_url + test_file.name, self.dst_url( ) ) )
            self._assert_key( test_file )
            buffer_size = s3am.operations.verify_buffer_size
            for verify_part_size in { buffer_size, part_size }:
                md5 = s3am.cli.main( concat(
                    'verify',
                    '--part-size', str( verify_part_size ),
                    self.dst_url( file_name=test_file.name ) ) )
                self.assertEquals( test_file.md5, md5 )


class UnreliableHandler( pyftpdlib.handlers.DTPHandler ):
    """
    Lets us trigger an IO error during the download
    """

    def send( self, data ):
        self._simulate_error( data )
        return pyftpdlib.handlers.DTPHandler.send( self, data )

    lock = Lock( )
    error_at_byte = None
    sent_bytes = 0

    @classmethod
    def _simulate_error( cls, data ):
        with cls.lock:
            if cls.error_at_byte is not None:
                cls.sent_bytes += len( data )
                if cls.sent_bytes > cls.error_at_byte:
                    log.info( 'Simulating error at %i', cls.sent_bytes )
                    cls.error_at_byte = None
                    cls.sent_bytes = 0
                    raise socket.error( )
                else:
                    log.info( 'Not simulating error at %i', cls.sent_bytes )

    @classmethod
    def setup_for_failure_at( cls, offset ):
        with cls.lock:
            cls.error_at_byte = offset
            cls.sent_bytes = 0



# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from setuptools import setup, find_packages

assert sys.version_info >= (2, 7)

from version import version

kwargs = dict(
    name='s3am',
    version=version,
    author='Hannes Schmidt',
    author_email='hannes@ucsc.edu',
    url='https://github.com/BD2KGenomics/cgcloud',
    description='Efficiently transfer large amounts of data to S3',
    package_dir={ '': 'src' },
    packages=find_packages( 'src' ),
    entry_points={
        'console_scripts': [
            's3am = s3am.cli:try_main'
        ]
    },
    install_requires=[
        'pycurl>=7.19.3',
        'boto==2.38.0',
        'bd2k-python-lib==1.13.dev16' ],
    tests_require=[
        'pytest==2.8.5',
        'pyftpdlib==1.5.0' ] )

from setuptools.command.test import test as TestCommand


class PyTest( TestCommand ):
    user_options = [ ('pytest-args=', 'a', "Arguments to pass to py.test") ]

    def initialize_options( self ):
        TestCommand.initialize_options( self )
        self.pytest_args = [ ]

    def finalize_options( self ):
        TestCommand.finalize_options( self )
        self.test_args = [ ]
        self.test_suite = True

    def run_tests( self ):
        import pytest
        # Sanitize command line arguments to avoid confusing Toil code attempting to parse them
        sys.argv[ 1: ] = [ ]
        errno = pytest.main( self.pytest_args )
        sys.exit( errno )


kwargs[ 'cmdclass' ] = { 'test': PyTest }

setup( **kwargs )
